@@ -18,13 +18,13 @@
 
 #include "argprocessing.hpp"
 
-#include <ccache/args.hpp>
 #include <ccache/argsinfo.hpp>
 #include <ccache/compopt.hpp>
 #include <ccache/context.hpp>
 #include <ccache/core/common.hpp>
 #include <ccache/depfile.hpp>
 #include <ccache/language.hpp>
+#include <ccache/util/args.hpp>
 #include <ccache/util/assertions.hpp>
 #include <ccache/util/direntry.hpp>
 #include <ccache/util/filesystem.hpp>
@@ -69,10 +69,11 @@ enum class OutputDepOrigin : uint8_t {
   wp = 2
 };
 
-struct ArgumentProcessingState
+class ArgumentProcessingState
 {
-  bool found_c_opt = false;
-  bool found_dc_opt = false;
+public:
+  std::optional<std::string> found_c_opt;
+  std::optional<std::string> found_dc_opt;
   bool found_S_opt = false;
   bool found_analyze_opt = false;
   bool found_pch = false;
@@ -83,8 +84,6 @@ struct ArgumentProcessingState
   bool found_valid_Fp = false;
   bool found_syntax_only = false;
   ColorDiagnostics color_diagnostics = ColorDiagnostics::automatic;
-  bool found_directives_only = false;
-  bool found_rewrite_includes = false;
   std::unordered_map<std::string, std::vector<std::string>> xarch_args;
   bool found_mf_opt = false;
   bool found_wp_md_or_mmd_opt = false;
@@ -104,36 +103,66 @@ struct ArgumentProcessingState
   // Arguments classified as input files.
   std::vector<fs::path> input_files;
 
-  // common_args contains all original arguments except:
-  // * those that never should be passed to the preprocessor,
-  // * those that only should be passed to the preprocessor (if run_second_cpp
-  //   is false), and
-  // * dependency options (like -MD and friends).
-  Args common_args;
-
-  // cpp_args contains arguments that were not added to common_args, i.e. those
-  // that should only be passed to the preprocessor if run_second_cpp is false.
-  // If run_second_cpp is true, they will be passed to the compiler as well.
-  Args cpp_args;
-
-  // dep_args contains dependency options like -MD. They are only passed to the
-  // preprocessor, never to the compiler.
-  Args dep_args;
-
-  // compiler_only_args contains arguments that should only be passed to the
-  // compiler, not the preprocessor.
-  Args compiler_only_args;
-
-  // compiler_only_args_no_hash contains arguments that should only be passed to
-  // the compiler, not the preprocessor, and that also should not be part of the
-  // hash identifying the result.
-  Args compiler_only_args_no_hash;
-
   // Whether to include the full command line in the hash.
   bool hash_full_command_line = false;
 
   // Whether to include the actual CWD in the hash.
   bool hash_actual_cwd = false;
+
+  template<typename T>
+  void
+  add_common_arg(T&& arg)
+  {
+    m_preprocessor_args.push_back(std::forward<T>(arg));
+    m_compiler_args.push_back(std::forward<T>(arg));
+  }
+
+  template<typename T>
+  void
+  add_compiler_only_arg(T&& arg)
+  {
+    m_compiler_args.push_back(std::forward<T>(arg));
+    m_extra_args_to_hash.push_back(std::forward<T>(arg));
+  }
+
+  template<typename T>
+  void
+  add_compiler_only_arg_no_hash(T&& arg)
+  {
+    m_compiler_args.push_back(std::forward<T>(arg));
+  }
+
+  template<typename T>
+  void
+  add_extra_args_to_hash(T&& args)
+  {
+    m_extra_args_to_hash.push_back(std::forward<T>(args));
+  }
+
+  template<typename T>
+  void
+  add_native_arg(T&& arg)
+  {
+    m_native_args.push_back(std::forward<T>(arg));
+  }
+
+  ProcessArgsResult
+  to_result()
+  {
+    return {
+      m_preprocessor_args,
+      m_compiler_args,
+      m_extra_args_to_hash,
+      m_native_args,
+      hash_actual_cwd,
+    };
+  }
+
+private:
+  util::Args m_preprocessor_args;
+  util::Args m_compiler_args;
+  util::Args m_extra_args_to_hash;
+  util::Args m_native_args;
 };
 
 bool
@@ -343,7 +372,7 @@ std::optional<Statistic>
 process_option_arg(const Context& ctx,
                    ArgsInfo& args_info,
                    Config& config,
-                   Args& args,
+                   util::Args& args,
                    size_t& args_index,
                    ArgumentProcessingState& state)
 {
@@ -351,7 +380,7 @@ process_option_arg(const Context& ctx,
 
   if (option_should_be_ignored(args[i], ctx.ignore_options())) {
     LOG("Not processing ignored option: {}", args[i]);
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -361,7 +390,7 @@ process_option_arg(const Context& ctx,
       LOG_RAW("--ccache-skip lacks an argument");
       return Statistic::bad_compiler_arguments;
     }
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -400,7 +429,7 @@ process_option_arg(const Context& ctx,
       ++argpath;
     }
     auto file_args =
-      Args::from_response_file(argpath, config.response_file_format());
+      util::Args::from_response_file(argpath, config.response_file_format());
     if (!file_args) {
       LOG("Couldn't read arg file {}", argpath);
       return Statistic::bad_compiler_arguments;
@@ -423,8 +452,8 @@ process_option_arg(const Context& ctx,
     // Argument is a comma-separated list of files.
     auto paths = util::split_into_strings(args[i], ",");
     for (auto it = paths.rbegin(); it != paths.rend(); ++it) {
-      auto file_args =
-        Args::from_response_file(*it, Args::ResponseFileFormat::posix);
+      auto file_args = util::Args::from_response_file(
+        *it, util::Args::ResponseFileFormat::posix);
       if (!file_args) {
         LOG("Couldn't read CUDA options file {}", *it);
         return Statistic::bad_compiler_arguments;
@@ -438,7 +467,7 @@ process_option_arg(const Context& ctx,
 
   if (arg == "-fdump-ipa-clones") {
     args_info.generating_ipa_clones = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -480,6 +509,10 @@ process_option_arg(const Context& ctx,
     const auto arch = arg.substr(7);
     auto it = state.xarch_args.emplace(arch, std::vector<std::string>()).first;
     it->second.emplace_back(args[i + 1]);
+    if (arch == "host" || arch == "device") {
+      state.add_common_arg(args[i]);
+      state.add_common_arg(args[i + 1]);
+    }
     ++i;
     return Statistic::none;
   }
@@ -492,9 +525,6 @@ process_option_arg(const Context& ctx,
     }
     ++i;
     args_info.arch_args.emplace_back(args[i]);
-    if (args_info.arch_args.size() == 2) {
-      config.set_run_second_cpp(true);
-    }
     return Statistic::none;
   }
 
@@ -507,11 +537,9 @@ process_option_arg(const Context& ctx,
           || args[i + 1] == "-include" || args[i + 1] == "--include"
           || args[i + 1] == "-fno-pch-timestamp")) {
     if (compopt_affects_compiler_output(args[i + 1])) {
-      state.compiler_only_args.push_back(args[i]);
-    } else if (compopt_affects_cpp_output(args[i + 1])) {
-      state.cpp_args.push_back(args[i]);
+      state.add_compiler_only_arg(args[i]);
     } else {
-      state.common_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
     }
     ++i;
     arg = make_dash_option(ctx.config, args[i]);
@@ -540,11 +568,11 @@ process_option_arg(const Context& ctx,
       || (i + 1 < args.size() && arg == "-Xclang"
           && compopt_affects_compiler_output(args[i + 1]))) {
     if (i + 1 < args.size() && arg == "-Xclang") {
-      state.compiler_only_args.push_back(args[i]);
+      state.add_compiler_only_arg(args[i]);
       ++i;
       arg = make_dash_option(ctx.config, args[i]);
     }
-    state.compiler_only_args.push_back(args[i]);
+    state.add_compiler_only_arg(args[i]);
     // Note: "-Xclang -option-that-takes-arg -Xclang arg" is not handled below
     // yet.
     if (compopt_takes_arg(arg)
@@ -553,7 +581,7 @@ process_option_arg(const Context& ctx,
         LOG("Missing argument to {}", args[i]);
         return Statistic::bad_compiler_arguments;
       }
-      state.compiler_only_args.push_back(args[i + 1]);
+      state.add_compiler_only_arg(args[i + 1]);
       ++i;
     }
     return Statistic::none;
@@ -562,10 +590,10 @@ process_option_arg(const Context& ctx,
       || (i + 1 < args.size() && arg == "-Xclang"
           && compopt_prefix_affects_compiler_output(args[i + 1]))) {
     if (i + 1 < args.size() && arg == "-Xclang") {
-      state.compiler_only_args.push_back(args[i]);
+      state.add_compiler_only_arg(args[i]);
       ++i;
     }
-    state.compiler_only_args.push_back(args[i]);
+    state.add_compiler_only_arg(args[i]);
     return Statistic::none;
   }
 
@@ -590,9 +618,8 @@ process_option_arg(const Context& ctx,
     }
   }
 
-  // We must have -c.
   if (arg == "-c" || arg == "--compile") { // --compile is NVCC
-    state.found_c_opt = true;
+    state.found_c_opt = args[i];
     return Statistic::none;
   }
 
@@ -613,34 +640,34 @@ process_option_arg(const Context& ctx,
 
     if (arg == "-TC") {
       args_info.actual_language = "c";
-      state.common_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       return Statistic::none;
     }
 
     if (arg == "-TP") {
       args_info.actual_language = "c++";
-      state.common_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       return Statistic::none;
     }
   }
 
-  // when using nvcc with separable compilation, -dc implies -c
+  // -dc implies -c when using NVCC with separable compilation.
   if ((arg == "-dc" || arg == "--device-c")
       && config.compiler_type() == CompilerType::nvcc) {
-    state.found_dc_opt = true;
+    state.found_dc_opt = args[i];
     return Statistic::none;
   }
 
   // -S changes the default extension.
   if (arg == "-S") {
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     state.found_S_opt = true;
     return Statistic::none;
   }
 
   // --analyze changes the default extension too
   if (arg == "--analyze") {
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     state.found_analyze_opt = true;
     return Statistic::none;
   }
@@ -651,7 +678,7 @@ process_option_arg(const Context& ctx,
       // an uppercase letter) is an ordinary Intel compiler option, not a
       // language specification. (GCC's "-x" language argument is always
       // lowercase.)
-      state.common_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       return Statistic::none;
     }
 
@@ -702,7 +729,7 @@ process_option_arg(const Context& ctx,
       || util::starts_with(arg, "-ffile-prefix-map=")) {
     std::string map = arg.substr(arg.find('=') + 1);
     args_info.debug_prefix_maps.push_back(map);
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -715,7 +742,7 @@ process_option_arg(const Context& ctx,
         LOG("Missing argument to {}", args[i]);
         return Statistic::bad_compiler_arguments;
       }
-      state.common_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       compilation_dir = args[i + 1];
       i++;
     } else {
@@ -725,14 +752,14 @@ process_option_arg(const Context& ctx,
       }
     }
     args_info.compilation_dir = std::move(compilation_dir);
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   // Debugging is handled specially, so that we know if we can strip line
   // number info.
   if (util::starts_with(arg, "-g")) {
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
 
     if (util::starts_with(arg, "-gdwarf")) {
       // Selection of DWARF format (-gdwarf or -gdwarf-<version>) enables
@@ -766,18 +793,18 @@ process_option_arg(const Context& ctx,
   if (config.is_compiler_group_msvc() && !config.is_compiler_group_clang()
       && is_msvc_z_debug_option(arg)) {
     state.last_seen_msvc_z_debug_option = args[i];
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   if (config.is_compiler_group_msvc() && util::starts_with(arg, "-Fd")) {
-    state.compiler_only_args_no_hash.push_back(args[i]);
+    state.add_compiler_only_arg_no_hash(args[i]);
     return Statistic::none;
   }
 
   if (config.is_compiler_group_msvc()
       && (util::starts_with(arg, "-MP") || arg == "-FS")) {
-    state.compiler_only_args_no_hash.push_back(args[i]);
+    state.add_compiler_only_arg_no_hash(args[i]);
     return Statistic::none;
   }
 
@@ -786,7 +813,7 @@ process_option_arg(const Context& ctx,
   if ((arg == "-MD" || arg == "-MMD") && !config.is_compiler_group_msvc()) {
     state.found_md_or_mmd_opt = true;
     args_info.generating_dependencies = true;
-    state.dep_args.push_back(args[i]);
+    state.add_compiler_only_arg(args[i]);
     return Statistic::none;
   }
 
@@ -814,10 +841,10 @@ process_option_arg(const Context& ctx,
     }
     // Keep the format of the args the same.
     if (separate_argument) {
-      state.dep_args.push_back("-MF");
-      state.dep_args.push_back(args_info.output_dep);
+      state.add_compiler_only_arg("-MF");
+      state.add_compiler_only_arg(args_info.output_dep);
     } else {
-      state.dep_args.push_back(FMT("-MF{}", args_info.output_dep));
+      state.add_compiler_only_arg(FMT("-MF{}", args_info.output_dep));
     }
     return Statistic::none;
   }
@@ -833,8 +860,8 @@ process_option_arg(const Context& ctx,
         LOG("Missing argument to {}", args[i]);
         return Statistic::bad_compiler_arguments;
       }
-      state.dep_args.push_back(args[i]);
-      state.dep_args.push_back(args[i + 1]);
+      state.add_compiler_only_arg(args[i]);
+      state.add_compiler_only_arg(args[i + 1]);
       dep_target = args[i + 1];
       i++;
     } else {
@@ -842,7 +869,7 @@ process_option_arg(const Context& ctx,
       const std::string_view arg_view(arg);
       const auto arg_opt = arg_view.substr(0, 3);
       dep_target = arg_view.substr(3);
-      state.dep_args.push_back(FMT("{}{}", arg_opt, dep_target));
+      state.add_compiler_only_arg(FMT("{}{}", arg_opt, dep_target));
     }
 
     if (args_info.dependency_target) {
@@ -851,7 +878,7 @@ process_option_arg(const Context& ctx,
       args_info.dependency_target = "";
     }
     *args_info.dependency_target +=
-      is_mq ? Depfile::escape_filename(dep_target) : dep_target;
+      is_mq ? depfile::escape_filename(dep_target) : dep_target;
 
     return Statistic::none;
   }
@@ -862,8 +889,7 @@ process_option_arg(const Context& ctx,
       && (util::starts_with(arg, "-MD") || util::starts_with(arg, "-MT")
           || util::starts_with(arg, "-LD"))) {
     // These affect compiler but also #define some things.
-    state.cpp_args.push_back(args[i]);
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -871,39 +897,39 @@ process_option_arg(const Context& ctx,
       // clang-cl:
       || arg == "-showIncludes:user") {
     args_info.generating_includes = true;
-    state.dep_args.push_back(args[i]);
+    state.add_compiler_only_arg(args[i]);
     return Statistic::none;
   }
 
   if (arg == "-fprofile-arcs") {
     args_info.profile_arcs = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   if (arg == "-ftest-coverage") {
     args_info.generating_coverage = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   if (arg == "-fstack-usage") {
     args_info.generating_stackusage = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   // This covers all the different marker cases
   if (util::starts_with(arg, "-fcallgraph-info")) {
     args_info.generating_callgraphinfo = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   // -Zs is MSVC's -fsyntax-only equivalent
   if (arg == "-fsyntax-only" || arg == "-Zs") {
     args_info.expect_output_obj = false;
-    state.compiler_only_args.push_back(args[i]);
+    state.add_compiler_only_arg(args[i]);
     state.found_syntax_only = true;
     return Statistic::none;
   }
@@ -912,7 +938,7 @@ process_option_arg(const Context& ctx,
       || arg == "-coverage") { // Undocumented but still works.
     args_info.profile_arcs = true;
     args_info.generating_coverage = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -922,7 +948,7 @@ process_option_arg(const Context& ctx,
       // the actual CWD in the .gcno file.
       state.hash_actual_cwd = true;
     }
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -933,20 +959,20 @@ process_option_arg(const Context& ctx,
       // The failure is logged by process_profiling_option.
       return Statistic::unsupported_compiler_option;
     }
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   if (util::starts_with(arg, "-fsanitize-blacklist=")) {
     args_info.sanitize_blacklists.emplace_back(args[i].substr(21));
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   if (util::starts_with(arg, "--sysroot=")) {
     auto path = std::string_view(arg).substr(10);
     auto relpath = core::make_relative_path(ctx, path);
-    state.common_args.push_back(FMT("--sysroot={}", relpath));
+    state.add_common_arg(FMT("--sysroot={}", relpath));
     return Statistic::none;
   }
 
@@ -956,9 +982,9 @@ process_option_arg(const Context& ctx,
       LOG("Missing argument to {}", args[i]);
       return Statistic::bad_compiler_arguments;
     }
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     auto relpath = core::make_relative_path(ctx, args[i + 1]);
-    state.common_args.push_back(relpath);
+    state.add_common_arg(relpath);
     i++;
     return Statistic::none;
   }
@@ -969,8 +995,8 @@ process_option_arg(const Context& ctx,
       LOG("Missing argument to {}", args[i]);
       return Statistic::bad_compiler_arguments;
     }
-    state.common_args.push_back(args[i]);
-    state.common_args.push_back(args[i + 1]);
+    state.add_common_arg(args[i]);
+    state.add_common_arg(args[i + 1]);
     i++;
     return Statistic::none;
   }
@@ -978,9 +1004,7 @@ process_option_arg(const Context& ctx,
   if (arg == "-P" || arg == "-Wp,-P") {
     // Avoid passing -P to the preprocessor since it removes preprocessor
     // information we need.
-    state.compiler_only_args.push_back(args[i]);
-    LOG("{} used; not compiling preprocessed code", args[i]);
-    config.set_run_second_cpp(true);
+    state.add_compiler_only_arg(args[i]);
     return Statistic::none;
   }
 
@@ -997,7 +1021,7 @@ process_option_arg(const Context& ctx,
         state.output_dep_origin = OutputDepOrigin::wp;
         args_info.output_dep = arg.substr(8);
       }
-      state.dep_args.push_back(args[i]);
+      state.add_compiler_only_arg(args[i]);
       return Statistic::none;
     } else if (util::starts_with(arg, "-Wp,-MMD,")
                && arg.find(',', 9) == std::string::npos) {
@@ -1007,19 +1031,19 @@ process_option_arg(const Context& ctx,
         state.output_dep_origin = OutputDepOrigin::wp;
         args_info.output_dep = arg.substr(9);
       }
-      state.dep_args.push_back(args[i]);
+      state.add_compiler_only_arg(args[i]);
       return Statistic::none;
     } else if ((util::starts_with(arg, "-Wp,-D")
                 || util::starts_with(arg, "-Wp,-U"))
                && arg.find(',', 6) == std::string::npos) {
-      state.cpp_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       return Statistic::none;
     } else if (arg == "-Wp,-MP"
                || (arg.size() > 8 && util::starts_with(arg, "-Wp,-M")
                    && arg[7] == ','
                    && (arg[6] == 'F' || arg[6] == 'Q' || arg[6] == 'T')
                    && arg.find(',', 8) == std::string::npos)) {
-      state.dep_args.push_back(args[i]);
+      state.add_compiler_only_arg(args[i]);
       return Statistic::none;
     } else if (config.direct_mode()) {
       // -Wp, can be used to pass too hard options to the preprocessor.
@@ -1029,12 +1053,12 @@ process_option_arg(const Context& ctx,
     }
 
     // Any other -Wp,* arguments are only relevant for the preprocessor.
-    state.cpp_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   if (arg == "-MP") {
-    state.dep_args.push_back(args[i]);
+    state.add_compiler_only_arg(args[i]);
     return Statistic::none;
   }
 
@@ -1058,16 +1082,16 @@ process_option_arg(const Context& ctx,
   if (config.compiler_type() == CompilerType::gcc) {
     if (arg == "-fdiagnostics-color" || arg == "-fdiagnostics-color=always") {
       state.color_diagnostics = ColorDiagnostics::always;
-      state.compiler_only_args_no_hash.push_back(args[i]);
+      state.add_compiler_only_arg_no_hash(args[i]);
       return Statistic::none;
     } else if (arg == "-fno-diagnostics-color"
                || arg == "-fdiagnostics-color=never") {
       state.color_diagnostics = ColorDiagnostics::never;
-      state.compiler_only_args_no_hash.push_back(args[i]);
+      state.add_compiler_only_arg_no_hash(args[i]);
       return Statistic::none;
     } else if (arg == "-fdiagnostics-color=auto") {
       state.color_diagnostics = ColorDiagnostics::automatic;
-      state.compiler_only_args_no_hash.push_back(args[i]);
+      state.add_compiler_only_arg_no_hash(args[i]);
       return Statistic::none;
     }
   } else if (config.is_compiler_group_clang()) {
@@ -1075,45 +1099,33 @@ process_option_arg(const Context& ctx,
     // -fcolor-diagnostics argument which is passed to cc1 is handled below.
     if (arg == "-Xclang" && i + 1 < args.size()
         && args[i + 1] == "-fcolor-diagnostics") {
-      state.compiler_only_args_no_hash.push_back(args[i]);
+      state.add_compiler_only_arg_no_hash(args[i]);
       ++i;
       arg = make_dash_option(ctx.config, args[i]);
     }
     if (arg == "-fdiagnostics-color" || arg == "-fdiagnostics-color=always"
         || arg == "-fcolor-diagnostics") {
       state.color_diagnostics = ColorDiagnostics::always;
-      state.compiler_only_args_no_hash.push_back(args[i]);
+      state.add_compiler_only_arg_no_hash(args[i]);
       return Statistic::none;
     } else if (arg == "-fno-diagnostics-color"
                || arg == "-fdiagnostics-color=never"
                || arg == "-fno-color-diagnostics") {
       state.color_diagnostics = ColorDiagnostics::never;
-      state.compiler_only_args_no_hash.push_back(args[i]);
+      state.add_compiler_only_arg_no_hash(args[i]);
       return Statistic::none;
     }
   }
 
-  // GCC
-  if (arg == "-fdirectives-only") {
-    state.found_directives_only = true;
-    return Statistic::none;
-  }
-
-  // Clang
-  if (arg == "-frewrite-includes") {
-    state.found_rewrite_includes = true;
-    return Statistic::none;
-  }
-
   if (arg == "-fno-pch-timestamp") {
     args_info.fno_pch_timestamp = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
   if (arg == "-fpch-preprocess") {
     state.found_fpch_preprocess = true;
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -1140,9 +1152,17 @@ process_option_arg(const Context& ctx,
       "Found -frecord-gcc-switches, hashing original command line unmodified");
   }
 
+  // -march=native, -mcpu=native and -mtune=native make the compiler optimize
+  // differently depending on platform.
+  if (arg == "-march=native" || arg == "-mcpu=native"
+      || arg == "-mtune=native") {
+    LOG("Detected system dependent argument: {}", args[i]);
+    state.add_native_arg(args[i]);
+  }
+
   // MSVC -u is something else than GCC -u, handle it specially.
   if (arg == "-u" && ctx.config.is_compiler_group_msvc()) {
-    state.cpp_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     return Statistic::none;
   }
 
@@ -1164,13 +1184,11 @@ process_option_arg(const Context& ctx,
     // rate. A secondary effect is that paths in the standard error output
     // produced by the compiler will be normalized.
     fs::path relpath = core::make_relative_path(ctx, args[i + next]);
-    auto& dest_args =
-      compopt_affects_cpp_output(arg) ? state.cpp_args : state.common_args;
-    dest_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
     if (next == 2) {
-      dest_args.push_back(args[i + 1]);
+      state.add_common_arg(args[i + 1]);
     }
-    dest_args.push_back(relpath);
+    state.add_common_arg(relpath);
 
     i += next;
     return Statistic::none;
@@ -1198,9 +1216,9 @@ process_option_arg(const Context& ctx,
       const auto relpath = core::make_relative_path(ctx, *path);
       std::string new_option = FMT("{}{}", option, relpath);
       if (compopt_affects_cpp_output(option)) {
-        state.cpp_args.push_back(std::move(new_option));
+        state.add_common_arg(std::move(new_option));
       } else {
-        state.common_args.push_back(std::move(new_option));
+        state.add_common_arg(std::move(new_option));
       }
       return Statistic::none;
     }
@@ -1214,11 +1232,11 @@ process_option_arg(const Context& ctx,
     }
 
     if (compopt_affects_cpp_output(arg)) {
-      state.cpp_args.push_back(args[i]);
-      state.cpp_args.push_back(args[i + 1]);
+      state.add_common_arg(args[i]);
+      state.add_common_arg(args[i + 1]);
     } else {
-      state.common_args.push_back(args[i]);
-      state.common_args.push_back(args[i + 1]);
+      state.add_common_arg(args[i]);
+      state.add_common_arg(args[i + 1]);
     }
 
     i++;
@@ -1234,14 +1252,14 @@ process_option_arg(const Context& ctx,
   if (arg[0] == '-') {
     if (compopt_affects_cpp_output(arg)
         || compopt_prefix_affects_cpp_output(arg)) {
-      state.cpp_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       return Statistic::none;
     } else if (ctx.config.is_compiler_group_msvc()
                && args[i][0] == '/' // Intentionally not checking arg here
                && DirEntry(args[i]).is_regular_file()) {
       // Likely the input file, which is handled in process_arg later.
     } else {
-      state.common_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       return Statistic::none;
     }
   }
@@ -1254,7 +1272,7 @@ Statistic
 process_arg(const Context& ctx,
             ArgsInfo& args_info,
             Config& config,
-            Args& args,
+            util::Args& args,
             size_t& args_index,
             ArgumentProcessingState& state)
 {
@@ -1275,7 +1293,7 @@ process_arg(const Context& ctx,
   if (!util::is_dev_null_path(args[i])) {
     if (!DirEntry(args[i]).is_regular_file()) {
       LOG("{} is not a regular file, not considering as input file", args[i]);
-      state.common_args.push_back(args[i]);
+      state.add_common_arg(args[i]);
       return Statistic::none;
     }
   }
@@ -1285,7 +1303,7 @@ process_arg(const Context& ctx,
     state.input_files.emplace_back(args[i]);
   } else {
     LOG("Not considering {} an input file since it doesn't exist", args[i]);
-    state.common_args.push_back(args[i]);
+    state.add_common_arg(args[i]);
   }
   return Statistic::none;
 }
@@ -1315,10 +1333,10 @@ process_args(Context& ctx)
   // args is a copy of the original arguments given to the compiler but with
   // arguments from @file and similar constructs expanded. It's only used as a
   // temporary data structure to loop over.
-  Args args = ctx.orig_args;
+  util::Args args = ctx.orig_args;
   ArgumentProcessingState state;
 
-  state.common_args.push_back(args[0]); // Compiler
+  state.add_common_arg(args[0]); // Compiler
 
   std::optional<Statistic> argument_error;
   for (size_t i = 1; i < args.size(); i++) {
@@ -1437,22 +1455,6 @@ process_args(Context& ctx)
     return tl::unexpected(*argument_error);
   }
 
-  if (state.generating_debuginfo_level_3 && !config.run_second_cpp()) {
-    // Debug level 3 makes line number information incorrect when compiling
-    // preprocessed code.
-    LOG_RAW("Generating debug info level 3; not compiling preprocessed code");
-    config.set_run_second_cpp(true);
-  }
-
-#ifdef __APPLE__
-  // Newer Clang versions on macOS are known to produce different debug
-  // information when compiling preprocessed code.
-  if (args_info.generating_debuginfo && !config.run_second_cpp()) {
-    LOG_RAW("Generating debug info; not compiling preprocessed code");
-    config.set_run_second_cpp(true);
-  }
-#endif
-
   if (state.found_pch || state.found_fpch_preprocess) {
     args_info.using_precompiled_header = true;
     if (!(config.sloppiness().contains(core::Sloppy::time_macros))) {
@@ -1464,12 +1466,6 @@ process_args(Context& ctx)
     }
   }
 
-  if (args_info.profile_generate && !config.run_second_cpp()) {
-    LOG_RAW(
-      "Generating profiling information; not compiling preprocessed code");
-    config.set_run_second_cpp(true);
-  }
-
   if (args_info.profile_path.empty()) {
     args_info.profile_path = ctx.apparent_cwd;
   }
@@ -1509,7 +1505,7 @@ process_args(Context& ctx)
 
   if (is_link) {
     if (args_info.output_is_precompiled_header) {
-      state.common_args.push_back("-c");
+      state.add_common_arg("-c");
     } else {
       LOG_RAW("No -c option found");
       // Having a separate statistic for autoconf tests is useful, as they are
@@ -1532,22 +1528,8 @@ process_args(Context& ctx)
     args_info.generating_dependencies = false;
   }
 
-  if (!config.run_second_cpp()
-      && (args_info.actual_language == "cu"
-          || args_info.actual_language == "cuda")) {
-    LOG("Source language is \"{}\"; not compiling preprocessed code",
-        args_info.actual_language);
-    config.set_run_second_cpp(true);
-  }
-
   args_info.direct_i_file = language_is_preprocessed(args_info.actual_language);
 
-  if (args_info.output_is_precompiled_header && !config.run_second_cpp()) {
-    // It doesn't work to create the .gch from preprocessed source.
-    LOG_RAW("Creating precompiled header; not compiling preprocessed code");
-    config.set_run_second_cpp(true);
-  }
-
   if (config.cpp_extension().empty()) {
     std::string p_language = p_language_for_language(args_info.actual_language);
     config.set_cpp_extension(extension_for_language(p_language).substr(1));
@@ -1587,14 +1569,14 @@ process_args(Context& ctx)
   // -finput-charset=CHARSET (otherwise conversion happens twice)
   // -x CHARSET (otherwise the wrong language is selected)
   if (!state.input_charset_option.empty()) {
-    state.cpp_args.push_back(state.input_charset_option);
+    state.add_common_arg(state.input_charset_option);
   }
   if (state.found_pch && !ctx.config.is_compiler_group_msvc()) {
-    state.cpp_args.push_back("-fpch-preprocess");
+    state.add_common_arg("-fpch-preprocess");
   }
   if (!state.explicit_language.empty()) {
-    state.cpp_args.push_back("-x");
-    state.cpp_args.push_back(state.explicit_language);
+    state.add_common_arg("-x");
+    state.add_common_arg(state.explicit_language);
   }
 
   args_info.strip_diagnostics_colors =
@@ -1621,21 +1603,6 @@ process_args(Context& ctx)
   if (args_info.generating_dependencies) {
     if (state.output_dep_origin == OutputDepOrigin::none) {
       args_info.output_dep = util::with_extension(args_info.output_obj, ".d");
-      if (!config.run_second_cpp()) {
-        // If we're compiling preprocessed code we're sending dep_args to the
-        // preprocessor so we need to use -MF to write to the correct .d file
-        // location since the preprocessor doesn't know the final object path.
-        state.dep_args.push_back("-MF");
-        state.dep_args.push_back(args_info.output_dep);
-      }
-    }
-
-    if (!args_info.dependency_target && !config.run_second_cpp()) {
-      // If we're compiling preprocessed code we're sending dep_args to the
-      // preprocessor so we need to use -MQ to get the correct target object
-      // file in the .d file.
-      state.dep_args.push_back("-MQ");
-      state.dep_args.push_back(args_info.output_obj);
     }
 
     if (!args_info.dependency_target) {
@@ -1666,7 +1633,7 @@ process_args(Context& ctx)
       }
 
       args_info.dependency_target =
-        Depfile::escape_filename(util::pstr(dep_target).str());
+        depfile::escape_filename(util::pstr(dep_target).str());
     }
   }
 
@@ -1687,41 +1654,15 @@ process_args(Context& ctx)
       ctx, util::add_extension(args_info.orig_input_file, ".000i.ipa-clones"));
   }
 
-  Args compiler_args = state.common_args;
-  compiler_args.push_back(state.compiler_only_args_no_hash);
-  compiler_args.push_back(state.compiler_only_args);
-
-  if (config.run_second_cpp()) {
-    compiler_args.push_back(state.cpp_args);
-  } else if (state.found_directives_only || state.found_rewrite_includes) {
-    // Need to pass the macros and any other preprocessor directives again.
-    compiler_args.push_back(state.cpp_args);
-    if (state.found_directives_only) {
-      state.cpp_args.push_back("-fdirectives-only");
-      // The preprocessed source code still needs some more preprocessing.
-      compiler_args.push_back("-fpreprocessed");
-      compiler_args.push_back("-fdirectives-only");
+  if (state.xarch_args.size() > 1) {
+    if (state.xarch_args.find("host") != state.xarch_args.end()) {
+      LOG_RAW("-Xarch_host in combination with other -Xarch_* is too hard");
+      return tl::unexpected(Statistic::unsupported_compiler_option);
     }
-    if (state.found_rewrite_includes) {
-      state.cpp_args.push_back("-frewrite-includes");
-      // The preprocessed source code still needs some more preprocessing.
-      compiler_args.push_back("-x");
-      compiler_args.push_back(args_info.actual_language);
+    if (state.xarch_args.find("device") != state.xarch_args.end()) {
+      LOG_RAW("-Xarch_device in combination with other -Xarch_* is too hard");
+      return tl::unexpected(Statistic::unsupported_compiler_option);
     }
-  } else if (!state.explicit_language.empty()) {
-    // Workaround for a bug in Apple's patched distcc -- it doesn't properly
-    // reset the language specified with -x, so if -x is given, we have to
-    // specify the preprocessed language explicitly.
-    compiler_args.push_back("-x");
-    compiler_args.push_back(p_language_for_language(state.explicit_language));
-  }
-
-  if (state.found_c_opt) {
-    compiler_args.push_back("-c");
-  }
-
-  if (state.found_dc_opt) {
-    compiler_args.push_back("-dc");
   }
 
   if (!state.xarch_args.empty()) {
@@ -1734,65 +1675,44 @@ process_args(Context& ctx)
   }
 
   for (const auto& arch : args_info.arch_args) {
-    compiler_args.push_back("-arch");
-    compiler_args.push_back(arch);
+    state.add_compiler_only_arg_no_hash("-arch");
+    state.add_compiler_only_arg_no_hash(arch);
 
     auto it = args_info.xarch_args.find(arch);
     if (it != args_info.xarch_args.end()) {
       args_info.xarch_args.emplace(arch, it->second);
 
       for (const auto& xarch : it->second) {
-        compiler_args.push_back("-Xarch_" + arch);
-        compiler_args.push_back(xarch);
+        state.add_compiler_only_arg_no_hash("-Xarch_" + arch);
+        state.add_compiler_only_arg_no_hash(xarch);
       }
     }
   }
 
-  Args preprocessor_args = state.common_args;
-  preprocessor_args.push_back(state.cpp_args);
-
-  if (config.run_second_cpp()) {
-    // When not compiling the preprocessed source code, only pass dependency
-    // arguments to the compiler to avoid having to add -MQ, supporting e.g.
-    // EDG-based compilers which don't support -MQ.
-    compiler_args.push_back(state.dep_args);
-  } else {
-    // When compiling the preprocessed source code, pass dependency arguments to
-    // the preprocessor since the compiler doesn't produce a .d file when
-    // compiling preprocessed source code.
-    preprocessor_args.push_back(state.dep_args);
-  }
-
-  Args extra_args_to_hash = state.compiler_only_args;
-  if (config.run_second_cpp()) {
-    extra_args_to_hash.push_back(state.dep_args);
-  }
   if (state.hash_full_command_line) {
-    extra_args_to_hash.push_back(ctx.orig_args);
+    state.add_extra_args_to_hash(ctx.orig_args);
   }
 
   if (diagnostics_color_arg) {
-    compiler_args.push_back(*diagnostics_color_arg);
-    if (!config.run_second_cpp()) {
-      // If we're compiling preprocessed code we're keeping any warnings from
-      // the preprocessor, so we need to make sure that they are in color.
-      preprocessor_args.push_back(*diagnostics_color_arg);
-    }
+    state.add_compiler_only_arg_no_hash(*diagnostics_color_arg);
   }
 
   if (ctx.config.depend_mode() && !args_info.generating_includes
       && ctx.config.compiler_type() == CompilerType::msvc) {
     ctx.auto_depend_mode = true;
     args_info.generating_includes = true;
-    compiler_args.push_back("/showIncludes");
+    state.add_compiler_only_arg_no_hash("/showIncludes");
   }
 
-  return ProcessArgsResult{
-    preprocessor_args,
-    extra_args_to_hash,
-    compiler_args,
-    state.hash_actual_cwd,
-  };
+  if (state.found_c_opt) {
+    state.add_compiler_only_arg_no_hash(*state.found_c_opt);
+  }
+
+  if (state.found_dc_opt) {
+    state.add_compiler_only_arg_no_hash(*state.found_dc_opt);
+  }
+
+  return state.to_result();
 }
 
 bool
