@@ -1,4 +1,4 @@
-// Copyright (C) 2021-2024 Joel Rosdahl and other contributors
+// Copyright (C) 2021-2025 Joel Rosdahl and other contributors
 //
 // See doc/AUTHORS.adoc for a complete list of contributors.
 //
@@ -252,6 +252,12 @@ const StatisticsField k_statistics_fields[] = {
         "Unsupported environment variable",
         FLAG_UNCACHEABLE),
 
+  // Source file (or an included header) has unsupported encoding. ccache
+  // currently requires UTF-8-encoded source code for MSVC.
+  FIELD(unsupported_source_encoding,
+        "Unsupported source encoding",
+        FLAG_UNCACHEABLE),
+
   // A source language e.g. specified with -x was unsupported by ccache.
   FIELD(unsupported_source_language,
         "Unsupported source language",
@@ -274,7 +280,7 @@ format_timestamp(const util::TimePoint& value)
     const auto tm = util::localtime(value);
     char buffer[100] = "?";
     if (tm) {
-      strftime(buffer, sizeof(buffer), "%c", &*tm);
+      (void)strftime(buffer, sizeof(buffer), "%c", &*tm);
     }
     return buffer;
   }
