@@ -1,4 +1,4 @@
-// Copyright (C) 2011-2024 Joel Rosdahl and other contributors
+// Copyright (C) 2011-2025 Joel Rosdahl and other contributors
 //
 // See doc/AUTHORS.adoc for a complete list of contributors.
 //
@@ -22,6 +22,7 @@
 #include <ccache/core/exceptions.hpp>
 #include <ccache/util/environment.hpp>
 #include <ccache/util/file.hpp>
+#include <ccache/util/filesystem.hpp>
 #include <ccache/util/format.hpp>
 
 #include <doctest/doctest.h>
@@ -30,6 +31,8 @@
 #include <string>
 #include <vector>
 
+namespace fs = util::filesystem;
+
 using doctest::Approx;
 using TestUtil::TestContext;
 
@@ -39,7 +42,7 @@ TEST_CASE("Config: default values")
 {
   Config config;
 
-  CHECK(config.base_dir().empty());
+  CHECK(config.base_dirs().empty());
   CHECK(config.cache_dir().empty()); // Set later
   CHECK(config.compiler().empty());
   CHECK(config.compiler_check() == "mtime");
@@ -79,7 +82,6 @@ TEST_CASE("Config: default values")
   CHECK_FALSE(config.remote_only());
   CHECK(config.remote_storage().empty());
   CHECK_FALSE(config.reshare());
-  CHECK(config.run_second_cpp());
   CHECK(config.sloppiness().to_bitmask() == 0);
   CHECK(config.stats());
   CHECK(config.temporary_dir().empty()); // Set later
@@ -99,57 +101,59 @@ TEST_CASE("Config::update_from_file")
   std::string base_dir = FMT("C:/{0}/foo/{0}", user);
 #endif
 
-  util::write_file(
+  REQUIRE(util::write_file(
     "ccache.conf",
-    "base_dir = " + base_dir + "\n"
-    "cache_dir=\n"
-    "cache_dir = $USER$/${USER}/.ccache\n"
-    "\n"
-    "\n"
-    "  #A comment\n"
-    "\t compiler = foo\n"
-    "compiler_check = none\n"
-    "compiler_type = nvcc\n"
-    "compression=false\n"
-    "compression_level= 2\n"
-    "cpp_extension = .foo\n"
-    "debug_dir = $USER$/${USER}/.ccache_debug\n"
-    "debug_level = 2\n"
-    "depend_mode = true\n"
-    "direct_mode = false\n"
-    "disable = true\n"
-    "extra_files_to_hash = a:b c:$USER\n"
-    "file_clone = true\n"
-    "hard_link = true\n"
-    "hash_dir = false\n"
-    "ignore_headers_in_manifest = a:b/c\n"
-    "ignore_options = -a=* -b\n"
-    "inode_cache = false\n"
-    "keep_comments_cpp = true\n"
-    "log_file = $USER${USER} \n"
-    "max_files = 17\n"
-    "max_size = 123M\n"
-    "msvc_dep_prefix = Some other prefix:\n"
-    "path = $USER.x\n"
-    "pch_external_checksum = true\n"
-    "prefix_command = x$USER\n"
-    "prefix_command_cpp = y\n"
-    "read_only = true\n"
-    "read_only_direct = true\n"
-    "recache = true\n"
-    "reshare = true\n"
-    "run_second_cpp = false\n"
-    "sloppiness =     time_macros   ,include_file_mtime"
-    "  include_file_ctime,file_stat_matches,file_stat_matches_ctime,pch_defines"
-    " ,  no_system_headers,system_headers,clang_index_store,ivfsoverlay,"
-    " gcno_cwd,\n"
-    "stats = false\n"
-    "temporary_dir = ${USER}_foo\n"
-    "umask = 777"); // Note: no newline.
+    "base_dir = " + base_dir
+      + "\n"
+        "cache_dir=\n"
+        "cache_dir = $USER$/${USER}/.ccache\n"
+        "\n"
+        "\n"
+        "  #A comment\n"
+        "\t compiler = foo\n"
+        "compiler_check = none\n"
+        "compiler_type = nvcc\n"
+        "compression=false\n"
+        "compression_level= 2\n"
+        "cpp_extension = .foo\n"
+        "debug_dir = $USER$/${USER}/.ccache_debug\n"
+        "debug_level = 2\n"
+        "depend_mode = true\n"
+        "direct_mode = false\n"
+        "disable = true\n"
+        "extra_files_to_hash = a:b c:$USER\n"
+        "file_clone = true\n"
+        "hard_link = true\n"
+        "hash_dir = false\n"
+        "ignore_headers_in_manifest = a:b/c\n"
+        "ignore_options = -a=* -b\n"
+        "inode_cache = false\n"
+        "keep_comments_cpp = true\n"
+        "log_file = $USER${USER} \n"
+        "max_files = 17\n"
+        "max_size = 123M\n"
+        "msvc_dep_prefix = Some other prefix:\n"
+        "path = $USER.x\n"
+        "pch_external_checksum = true\n"
+        "prefix_command = x$USER\n"
+        "prefix_command_cpp = y\n"
+        "read_only = true\n"
+        "read_only_direct = true\n"
+        "recache = true\n"
+        "reshare = true\n"
+        "sloppiness =     time_macros   ,include_file_mtime"
+        "  "
+        "include_file_ctime,file_stat_matches,file_stat_matches_ctime,pch_"
+        "defines"
+        " ,  no_system_headers,system_headers,clang_index_store,ivfsoverlay,"
+        " gcno_cwd,\n"
+        "stats = false\n"
+        "temporary_dir = ${USER}_foo\n"
+        "umask = 777")); // Note: no newline.
 
   Config config;
   REQUIRE(config.update_from_file("ccache.conf"));
-  CHECK(config.base_dir() == base_dir);
+  CHECK(config.base_dirs() == std::vector<fs::path>{base_dir});
   CHECK(config.cache_dir() == FMT("{0}$/{0}/.ccache", user));
   CHECK(config.compiler() == "foo");
   CHECK(config.compiler_check() == "none");
@@ -182,7 +186,6 @@ TEST_CASE("Config::update_from_file")
   CHECK(config.read_only_direct());
   CHECK(config.recache());
   CHECK(config.reshare());
-  CHECK_FALSE(config.run_second_cpp());
   CHECK(config.sloppiness().to_bitmask()
         == (static_cast<uint32_t>(core::Sloppy::clang_index_store)
             | static_cast<uint32_t>(core::Sloppy::file_stat_matches)
@@ -207,31 +210,31 @@ TEST_CASE("Config::update_from_file, error handling")
 
   SUBCASE("missing equal sign")
   {
-    util::write_file("ccache.conf", "no equal sign");
+    REQUIRE(util::write_file("ccache.conf", "no equal sign"));
     REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
                         "ccache.conf:1: missing equal sign");
   }
 
   SUBCASE("unknown key")
   {
-    util::write_file("ccache.conf", "# Comment\nfoo = bar");
+    REQUIRE(util::write_file("ccache.conf", "# Comment\nfoo = bar"));
     CHECK(config.update_from_file("ccache.conf"));
   }
 
   SUBCASE("invalid bool")
   {
-    util::write_file("ccache.conf", "disable=");
+    REQUIRE(util::write_file("ccache.conf", "disable="));
     REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
                         "ccache.conf:1: not a boolean value: \"\"");
 
-    util::write_file("ccache.conf", "disable=foo");
+    REQUIRE(util::write_file("ccache.conf", "disable=foo"));
     REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
                         "ccache.conf:1: not a boolean value: \"foo\"");
   }
 
   SUBCASE("invalid variable reference")
   {
-    util::write_file("ccache.conf", "base_dir = ${foo");
+    REQUIRE(util::write_file("ccache.conf", "base_dir = ${foo"));
     REQUIRE_THROWS_WITH(
       config.update_from_file("ccache.conf"),
       "ccache.conf:1: syntax error: missing '}' after \"foo\"");
@@ -240,14 +243,14 @@ TEST_CASE("Config::update_from_file, error handling")
 
   SUBCASE("empty umask")
   {
-    util::write_file("ccache.conf", "umask = ");
+    REQUIRE(util::write_file("ccache.conf", "umask = "));
     CHECK(config.update_from_file("ccache.conf"));
     CHECK(config.umask() == std::nullopt);
   }
 
   SUBCASE("invalid size")
   {
-    util::write_file("ccache.conf", "max_size = foo");
+    REQUIRE(util::write_file("ccache.conf", "max_size = foo"));
     REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
                         "ccache.conf:1: invalid size: \"foo\"");
     // Other cases tested in test_Util.c.
@@ -255,7 +258,7 @@ TEST_CASE("Config::update_from_file, error handling")
 
   SUBCASE("unknown sloppiness")
   {
-    util::write_file("ccache.conf", "sloppiness = time_macros, foo");
+    REQUIRE(util::write_file("ccache.conf", "sloppiness = time_macros, foo"));
     CHECK(config.update_from_file("ccache.conf"));
     CHECK(config.sloppiness().to_bitmask()
           == static_cast<uint32_t>(core::Sloppy::time_macros));
@@ -263,15 +266,15 @@ TEST_CASE("Config::update_from_file, error handling")
 
   SUBCASE("invalid unsigned")
   {
-    util::write_file("ccache.conf", "max_files =");
+    REQUIRE(util::write_file("ccache.conf", "max_files ="));
     REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
                         "ccache.conf:1: invalid unsigned integer: \"\"");
 
-    util::write_file("ccache.conf", "max_files = -42");
+    REQUIRE(util::write_file("ccache.conf", "max_files = -42"));
     REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
                         "ccache.conf:1: invalid unsigned integer: \"-42\"");
 
-    util::write_file("ccache.conf", "max_files = foo");
+    REQUIRE(util::write_file("ccache.conf", "max_files = foo"));
     REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
                         "ccache.conf:1: invalid unsigned integer: \"foo\"");
   }
@@ -283,12 +286,11 @@ TEST_CASE("Config::update_from_file, error handling")
 
   SUBCASE("relative base dir")
   {
-    util::write_file("ccache.conf", "base_dir = relative/path");
-    REQUIRE_THROWS_WITH(
-      config.update_from_file("ccache.conf"),
-      "ccache.conf:1: not an absolute path: \"relative/path\"");
+    REQUIRE(util::write_file("ccache.conf", "base_dir = relative"));
+    REQUIRE_THROWS_WITH(config.update_from_file("ccache.conf"),
+                        "ccache.conf:1: not an absolute path: \"relative\"");
 
-    util::write_file("ccache.conf", "base_dir =");
+    REQUIRE(util::write_file("ccache.conf", "base_dir ="));
     CHECK(config.update_from_file("ccache.conf"));
   }
 }
@@ -310,13 +312,13 @@ TEST_CASE("Config::update_from_environment")
 
 TEST_CASE("Config::response_file_format")
 {
-  using ResponseFileFormat = Args::ResponseFileFormat;
+  using ResponseFileFormat = util::Args::ResponseFileFormat;
 
   Config config;
 
   SUBCASE("from config gcc")
   {
-    util::write_file("ccache.conf", "response_file_format = posix");
+    REQUIRE(util::write_file("ccache.conf", "response_file_format = posix"));
     CHECK(config.update_from_file("ccache.conf"));
 
     CHECK(config.response_file_format() == ResponseFileFormat::posix);
@@ -324,7 +326,7 @@ TEST_CASE("Config::response_file_format")
 
   SUBCASE("from config msvc")
   {
-    util::write_file("ccache.conf", "response_file_format = windows");
+    REQUIRE(util::write_file("ccache.conf", "response_file_format = windows"));
     CHECK(config.update_from_file("ccache.conf"));
 
     CHECK(config.response_file_format() == ResponseFileFormat::windows);
@@ -332,8 +334,8 @@ TEST_CASE("Config::response_file_format")
 
   SUBCASE("from config msvc with clang compiler")
   {
-    util::write_file("ccache.conf",
-                     "response_file_format = windows\ncompiler_type = clang");
+    REQUIRE(util::write_file(
+      "ccache.conf", "response_file_format = windows\ncompiler_type = clang"));
     CHECK(config.update_from_file("ccache.conf"));
 
     CHECK(config.response_file_format() == ResponseFileFormat::windows);
@@ -341,7 +343,7 @@ TEST_CASE("Config::response_file_format")
 
   SUBCASE("guess from compiler gcc")
   {
-    util::write_file("ccache.conf", "compiler_type = clang");
+    REQUIRE(util::write_file("ccache.conf", "compiler_type = clang"));
     CHECK(config.update_from_file("ccache.conf"));
 
     CHECK(config.response_file_format() == ResponseFileFormat::posix);
@@ -349,7 +351,7 @@ TEST_CASE("Config::response_file_format")
 
   SUBCASE("guess from compiler msvc")
   {
-    util::write_file("ccache.conf", "compiler_type = msvc");
+    REQUIRE(util::write_file("ccache.conf", "compiler_type = msvc"));
     CHECK(config.update_from_file("ccache.conf"));
 
     CHECK(config.response_file_format() == ResponseFileFormat::windows);
@@ -363,7 +365,7 @@ TEST_CASE("Config::set_value_in_file")
 
   SUBCASE("set new value")
   {
-    util::write_file("ccache.conf", "path = vanilla\n");
+    REQUIRE(util::write_file("ccache.conf", "path = vanilla\n"));
     config.set_value_in_file("ccache.conf", "compiler", "chocolate");
     std::string content = *util::read_file<std::string>("ccache.conf");
     CHECK(content == "path = vanilla\ncompiler = chocolate\n");
@@ -371,7 +373,8 @@ TEST_CASE("Config::set_value_in_file")
 
   SUBCASE("existing value")
   {
-    util::write_file("ccache.conf", "path = chocolate\nstats = chocolate\n");
+    REQUIRE(
+      util::write_file("ccache.conf", "path = chocolate\nstats = chocolate\n"));
     config.set_value_in_file("ccache.conf", "path", "vanilla");
     std::string content = *util::read_file<std::string>("ccache.conf");
     CHECK(content == "path = vanilla\nstats = chocolate\n");
@@ -379,7 +382,8 @@ TEST_CASE("Config::set_value_in_file")
 
   SUBCASE("unknown option")
   {
-    util::write_file("ccache.conf", "path = chocolate\nstats = chocolate\n");
+    REQUIRE(
+      util::write_file("ccache.conf", "path = chocolate\nstats = chocolate\n"));
     try {
       config.set_value_in_file("ccache.conf", "foo", "bar");
       CHECK(false);
@@ -393,7 +397,7 @@ TEST_CASE("Config::set_value_in_file")
 
   SUBCASE("unknown sloppiness")
   {
-    util::write_file("ccache.conf", "path = vanilla\n");
+    REQUIRE(util::write_file("ccache.conf", "path = vanilla\n"));
     config.set_value_in_file("ccache.conf", "sloppiness", "foo");
     std::string content = *util::read_file<std::string>("ccache.conf");
     CHECK(content == "path = vanilla\nsloppiness = foo\n");
@@ -401,7 +405,7 @@ TEST_CASE("Config::set_value_in_file")
 
   SUBCASE("comments are kept")
   {
-    util::write_file("ccache.conf", "# c1\npath = blueberry\n#c2\n");
+    REQUIRE(util::write_file("ccache.conf", "# c1\npath = blueberry\n#c2\n"));
     config.set_value_in_file("ccache.conf", "path", "vanilla");
     config.set_value_in_file("ccache.conf", "compiler", "chocolate");
     std::string content = *util::read_file<std::string>("ccache.conf");
@@ -434,14 +438,16 @@ TEST_CASE("Config::visit_items")
 {
   TestContext test_context;
 
-  util::write_file(
+#ifndef _WIN32
+#  define BASE_DIR "/bd\n"
+#else
+#  define BASE_DIR "C:\\bd\n"
+#endif
+
+  REQUIRE(util::write_file(
     "test.conf",
     "absolute_paths_in_stderr = true\n"
-#ifndef _WIN32
-    "base_dir = /bd\n"
-#else
-    "base_dir = C:\\bd\n"
-#endif
+    "base_dir = " BASE_DIR
     "cache_dir = cd\n"
     "compiler = c\n"
     "compiler_check = cc\n"
@@ -479,14 +485,14 @@ TEST_CASE("Config::visit_items")
     "remote_storage = rs\n"
     "reshare = true\n"
     "response_file_format = posix\n"
-    "run_second_cpp = false\n"
     "sloppiness = include_file_mtime, include_file_ctime, time_macros,"
     " file_stat_matches, file_stat_matches_ctime, pch_defines, system_headers,"
     " clang_index_store, ivfsoverlay, gcno_cwd \n"
     "stats = false\n"
     "stats_log = sl\n"
     "temporary_dir = td\n"
-    "umask = 022\n");
+    "umask = 022\n"));
+#undef BASE_DIR
 
   Config config;
   config.update_from_file("test.conf");
@@ -542,7 +548,6 @@ TEST_CASE("Config::visit_items")
     "(test.conf) remote_storage = rs",
     "(test.conf) reshare = true",
     "(test.conf) response_file_format = posix",
-    "(test.conf) run_second_cpp = false",
     "(test.conf) sloppiness = clang_index_store, file_stat_matches,"
     " file_stat_matches_ctime, gcno_cwd, include_file_ctime,"
     " include_file_mtime, ivfsoverlay, pch_defines, system_headers,"
